from TCPModemDriver import TCPModemDriver
from ModemDriver import ModemDriver
from SerialDriver import SerialDriver
from DeviceDriver import DeviceDriver
from Tunnel import Tunnel
from datetime import datetime
import logging
import serial
import os
import time
import threading



class AccessPoint():
    messageBuffer = b""
    DIALSEND = 20001
    DIALRECV = 20002
    LISTENSEND = 20002
    LISTENRECV = 20001
    
    
    waitThread:threading.Thread
    deviceDriver : DeviceDriver = None
    tunnel = Tunnel()
    killWait = False
    t1 : threading.Thread
    t2 : threading.Thread
    logger = logging.getLogger('dreampi')
    state = "starting"

    def assignCommPort(self,com_port) -> str:
        self.deviceDriver = None
        try:
            self.deviceDriver = ModemDriver(com_port,False)           
            if self.deviceDriver.modem is not None:
                self.com_port = com_port
                return self.com_port
        except serial.SerialException as e:
            message = str(e)
            # print(message)
            #if "could not open port" in message:
            self.com_port = None
        except IOError as e:
            self.com_port = None
        finally:
            if self.deviceDriver is not None:
                self.deviceDriver.disconnect()
                self.deviceDriver = ModemDriver(com_port)
        
        try:
            print("Checking serial driver")
            self.deviceDriver = SerialDriver(com_port)
            if self.deviceDriver is not None and self.deviceDriver.serialDevice is not None:
                print("Serial Name")
                print(self.deviceDriver.serialDevice.name)
                self.com_port = com_port
                return self.com_port
        except serial.SerialException as e:
            message = str(e)
            print("AP Message %s" % message)
            if "could not open port" in message:
                self.com_port = None
        except IOError as e:
            self.com_port = None
         
        #assign to tcp
        self.com_port = "TCP"
        self.deviceDriver = TCPModemDriver()
        return self.com_port

    def scanCommPorts(self) -> str:
        for i in range(3,25): # this should be a big enough range. USB com ports usually end up in the teens.
            osName = os.name
            if osName == 'posix': # should work on linux and Mac for USB modem, but untested.
                com_port = "/dev/ttyACM%s" % i
            else:
                com_port = "COM%s" % i
                
            com_port = self.assignCommPort(com_port)
            
            if com_port != "TCP":
                return com_port
        
        return com_port
            
    def send_directory(self):
        with open('directory.txt') as f:
            lines = f.readlines()
            for element in lines:
                self.write(element.encode())
            
        self.write(b'\xFF')
        print("Sent")
        
    def send_playerInfo(self):
        with open('playerinfo.txt') as f:
            lines = f.readlines()
            for element in lines:
                self.write(element.encode())
            
        self.write(b'\xFF')
        print("Sent")

    def get_playerID(self) -> str:
        with open('playerinfo.txt') as f:
            lines = f.readlines()
            for element in lines:
                if "name:" in element:
                    return element[5:]
            
        return ""

    def send_test(self):
        self.write("%Anthony".encode())
        
    def disconnect(self):
        self.deviceDriver.disconnect()
        
    def ring(self):
        #self.logger.info("Sending Ring")
        self.write(b'%')
        
        #deviceDriver.writeByte(("RING\r\n").encode())
        #deviceDriver.writeByte.write(("CONNECT\r\n").encode())
        #self.logger.info("Ready for Data Exchange!")
        #tcp.shutdown(socket.SHUT_RDWR) #best practice is to close your socket, but it gives me issues.
        #tcp.close()
    def readDevice(self):
        waiting = self.deviceDriver.in_waiting()
        if waiting > 0:
            self.messageBuffer += self.deviceDriver.read(waiting)
        
        
    def readByte(self):
        self.readDevice()
        if len(self.messageBuffer) == 0:
            return None
        print("AP:readByte")
        b = self.messageBuffer[0:1]
        self.messageBuffer = self.messageBuffer[1:]
        return b
        
    def read(self):
        self.readDevice()
        if len(self.messageBuffer) == 0:
            return None
        #print("AP:read")
        b = self.messageBuffer[0:]
        self.messageBuffer = b""
        return b   
        
    def write(self,toSend):
        self.deviceDriver.writeByte(toSend)
        
    def listener(self):
        self.logger.info("listening")
        last = 0
        currentSequence = 0
        while(self.state != "netlink_disconnected"):
            packetSet = self.tunnel.recvUDP()
            
            if packetSet:
                packetSplit = b"<packetSplit>"
                dataSplit = b"<dataSplit>"
                #print(packetSet)print("Sending UDP packet")sendUdp
                packets= packetSet.split(packetSplit)
                try:
                    while True:
                        packetNum = 0
                        
                        #go through all packets 
                        for p in packets:
                          if int(p.split(dataSplit)[1]) == currentSequence:
                            break
                          packetNum += 1
                        
                        #if the packet needed is not here,  grab the latest in the set
                        if packetNum == len(packets):
                            packetNum = 0
                        
                        message = packets[packetNum]
                        payload = message.split(dataSplit)[0]
                        sequence = message.split(dataSplit)[1]
                        if int(sequence) < currentSequence:
                            break  #All packets are old data, so drop it entirely
                        
                        currentSequence = int(sequence) + 1
                        
                        toSend = payload
                        
                        self.write(toSend)
                        if len(payload) > 0:
                            self.logger.info("AP: (listener) Payload %s" % payload)
                        if packetNum == 0: # if the first packet was the processed packet,  no need to go through the rest
                            break

                except IndexError:
                    continue
                    
        self.logger.info("listener stopped")        
                
    def sender(self,oppPort):
        self.logger.info("sending")
        last = 0
        sequence = 0
        packets = []
        packetSplit = b"<packetSplit>"
        dataSplit = b"<dataSplit>"
        while(self.state != "netlink_disconnected"):
            raw_input = self.read()
           
            if raw_input is None:
                continue
            
            if b"NO CARRIER" in raw_input:
                self.state = "netlink_disconnected"
                time.sleep(1)
                self.tunnel.closeUDP()
                self.logger.info("sender stopped")
                return
            
            try:
                payload = raw_input
                
                seq = str(sequence)
                if len(payload)>0:
                    #print("AP: Payload (sender) %s" % payload)
                    packets.insert(0,(payload+dataSplit+seq.encode()))
                    if(len(packets) > 5):
                        packets.pop()
                    #print("Sending UDP packet")
                    #print("Sending UDP packet %s" % packetSplit.join(packets))
                    self.tunnel.sendUDP(packetSplit.join(packets),oppPort)            
                    sequence+=1
            except Exception as error:
                print(error)
                continue

    def approved(self):
        print("Sending ready")
        if self.tunnel.ready():
            self.write(b'*')
           
            print("Starting threads")
            self.t1 = threading.Thread(target=self.listener)
            print("Thread 1")
            self.t2 = threading.Thread(target=self.sender,args=[self.DIALSEND])
            print("Thread 2")
            self.tunnel.establishUDPConnection(self.DIALRECV)
            print("Establish")
            self.t1.start()
            self.t2.start()
            self.t1.join()
            self.t2.join()
        while(self.t1):
            pass

    def dial(self):
        self.killWait= True
        ipAddress = b""
        char = self.readByte()
        
        while char != b'#':
            if char:
                ipAddress += char
                
            char = self.readByte()
        
        print("Dialing %s" % ipAddress.decode())
        
        
        try:
            while self.readByte() != None:
                pass
        finally:
            pass
      
        print("Initializing tunnel connection")
        self.state = self.tunnel.initConnection(ipAddress.decode(),self.get_playerID())
        print("Initialized")
        
        
        if self.state == "failed":
            print(self.state)
            self.disconnect()
            print("Hanging up")
            return None
            
        self.write(b'#')

                    
    def do_routing(self):
        ipAddress = b""
        try:
            now = datetime.now()
            while (datetime.now() - now).total_seconds() < 360:
                char = self.readByte()
                
                if not char:
                    continue
                print("Routing: %s" % char)
                now = datetime.now() 
                
                #used when netlinkOS is sending opponent info
                if char == b'@':
                    self.send_test()
                    
                #used when netlinkOS needs a test
                if char == b'^':
                    self.send_test()
                
                #used when netlinkOS requests player info
                if char == b'&':
                    self.send_playerInfo()
                
                #used when netlinkOS requests directory
                if char == b'%':
                    self.send_directory()
                    
                #used when netlinkOS is dialing   
                if char == b'#':
                    print("Dial command")
                    self.dial()
                
                #used when netlinkOS sends acknowledge  
                if char == b'$':
                    print("Sending acknowedge");
                    self.tunnel.acknowledge()
                    print("Done")
                    # At this point, waitThread is in control
                    while(self.waitThread):
                        pass
                    return;
                    
                #used when netlinkOS waits for acknowledge  
                if char == b':':
                    print("Checking ack")
                    if self.tunnel.checkAcknowledge():
                        print("Approved")
                        self.approved()
                    while(self.waitThread):
                        pass 
                    return;
                
            print("Timed out")
                
               

        except IOError as error:
            print(error)
            ex_type, ex_value, ex_traceback = sys.exc_info()
            
            print("Exception type : %s" % ex_type.__name__)
            print("Exception message : %s" %ex_value)
        except Exception as error:
            print(error)
        finally:
            print("Done");
            
            return

    def wait(self):  
        self.killWait = False
        self.waitThread = threading.Thread(target=self.beginWaiting, daemon=True)
        self.waitThread.start()
        
    
    def beginWaiting(self):
        self.tunnel.listening()
        while self.tunnel.accept() == False and self.killWait == False:
            continue;
            
        if self.killWait:
            return
        
       
        while True:
            try:
                data = self.tunnel.messages()
            except socket.error: #first try can return no payload
                continue
            if "name:" in data.decode():
                self.write(b'%')
                self.write(data.decode().replace("name:","").encode())
                continue
                
            if data == b'readyip':
                self.tunnel.begin()
                self.ring()
                print("Starting threads")
                self.t1 = threading.Thread(target=self.listener)
                print("Thread 1")
                self.t2 = threading.Thread(target=self.sender,args=[self.LISTENSEND])
                print("Thread 2")
                self.tunnel.establishUDPConnection(self.LISTENRECV)
                print("Establish")
                self.t1.start()
                self.t2.start()
                self.t1.join()
                self.t2.join()
                while(self.t1):
                    pass
                return 
            if not data:
                print("failed to init")


    def connect(self) -> bool:
        
        isConnected = self.deviceDriver.connect()
        if not isConnected:
            return isConnected
            
        self.deviceDriver.netlinkAnswered()
        time.sleep(5)
        print("Sending response")
        self.write(b'*');
        
        print("Starting tunnel wait thread")
        self.wait()
        print("AccessPoint connected to device")
        return isConnected
              