import time
import serial
from datetime import datetime
from modemClass import Modem

class ModemDriver:
    serialDevice: serial.Serial
    waitTime = 8.0
    modem: Modem
    
    def __init__(self, com_port: str, sendTone:bool=True):
       #110, 300, 1200, 2400, 4800, 9600, 19200, 38400, 57600, or 115200
        speed =  115200 #209954 
 
        self.modem = Modem(com_port, speed,send_dial_tone=sendTone)
        #self.modem.connect_netlink()
        self.modem.connect_netlink(speed=57600,timeout=0.01,rtscts=True) #non-blocking version
        #print("potential modem found at %s" % com)
        self.modem.query_modem("AT",timeout = 1) # potential modem. Other devices respond to AT, so not definitive.
        self.modem.query_modem("AT+FCLASS=8",timeout = 1) # if potential modem, find out if it's our voice modem
        self.modem.reset()
   
    def in_waiting(self) : return self.serialDevice.in_waiting
    
    def read(self,length ) -> bytearray:
        data = self.serialDevice.read(length)
        #char = data.strip().decode()
        if len(data) == 0:
            return None
        return data
   
        
    def readByte(self) -> bytearray:
        data = self.serialDevice.read()
        #char = data.strip().decode()
        if len(data) == 0:
            return None
        #print("Data", "0x" + format(ord(char), '02x') + ' ' + char)
        return data
        
    def writeByte(self,data: str):
        self.serialDevice.write(data)
        
    def listen(self):
        pass
        
    def connect(self):
        self.modem.connect()
        self.modem.send_escape()
        self.modem.start_dial_tone()
        self.serialDevice = self.modem._serial
        print("Flushing")
        self.serialDevice.flush()
        self.serialDevice.reset_input_buffer()
        self.serialDevice.reset_output_buffer()
        time.sleep(1)
        
        while True:
            char = self.readByte()
            if not char:
                break
        
        print("Waiting for device response")
        while True:
            self.update()
            
            char = self.readByte()
            
            if not char:
                continue
            print("%s" % char)
            
            if ord(char) == 16:
                # DLE character
                try:
                    char = self.readByte()
      
                    if not char:
                        continue
                 
            
             
                    print("stopping dial tone")
                    self.modem.stop_dial_tone()
                    time_digit_heard = datetime.now()
                    while (datetime.now() - time_digit_heard).total_seconds() <= self.waitTime:
                        continue;
                        
                    print("Done waiting")
                    
                    time_digit_heard = None
                    self.modem.connect_netlink(speed=57600,timeout=0.01,rtscts=True) #non-blocking version
                
                
                
                    self.modem.query_modem(b'AT%E0')
                    self.modem.query_modem(b"AT\N3\V1%C0")
                    self.modem.query_modem(b'AT+MS=V32b,1,14400,14400,14400,14400')
                    self.modem.query_modem(b"ATA", timeout=120, response = "CONNECT")
                    self.serialDevice = self.modem._serial
                    print("Finished connecting")
                    return True
                except IOError as error:
                    print(IOError)
                    return False
                    
                except (TypeError, ValueError):
                    print(TypeError)
                    print(ValueError)
                    pass    


    
    def connectFinished(self):
        pass
    def hangup(self):
        try:
            self.modem.send_command('ATH0')
        except IOError as error:
            pass
        finally:
            pass
    def disconnect(self):
        try:     
            self.serialDevice = self.modem._serial
            self.serialDevice.flush()
            for i in range(3):
                self.writeByte(b'+')
                time.sleep(0.2)
            time.sleep(4)
            self.modem.disconnect()
            self.serialDevice = None
        except IOError as error:
            pass
        finally:
            pass
        
    def update(self):
        self.modem.update()
        
    def answered(self): 
        self.modem.answer()
        self.modem.disconnect()
        
    def netlinkAnswered(self):
        pass

        
           
    def device_status(self):
        print("parsing")
        while (True):
            data = self.readByte() #this is blocking, but the modem sends <DLE>s at regular intervals to indicate silence
            
            if not data: #which makes this pointless
                continue
            print("data found %s" % data)
            char = data.strip()
            if not char: #which makes this pointless
                continue
            char = char.decode()
            try:
                if char == '#':   
                    print("# found")                
                    self.sendByte("#")
                    print("done parsing")
                    return True
            except (TypeError, ValueError):#Dreampi originally tried to convert characters to int and passed on the exception raised for other characters. This shouldn't be needed anymore.
                pass        
        return False